import numpy as np
from numpy import linalg as la
from typing import Iterable, Optional
from compmec.nurbs import SplineBaseFunction
from compmec.nurbs.curves import SplineCurve

def transform2U(u: Iterable[float], n: float, p: float, algorithm: int = 1):
    U = np.zeros(n+p+1)
    if algorithm == 1:
        for j in range(1, n-p):
            U[j+p] = np.sum(u[j:j+p])/p
    else:
        ValueError("Algorithm is nod valid!")
    U[n:] = 1
    return U

def curve_control_points(L: np.ndarray, points: np.ndarray):
    """
    t is a vector of the position of each point.
        path(t[i]) = points[i]
    points is a 2D numpy matrix of shape (npts, dim)
    """
    n = L.shape[0]
    points = np.array(points).T
    npts, dim = points.shape
    controlpts = np.zeros((n, dim))
    for i in range(dim):
        controlpts[:, i] = compute_control_points(L, points[:, i])
    return controlpts

def curve_spline(t: Iterable[float], points: np.ndarray, p: Optional[int] = None, n: Optional[int] = None):
    t = (t - min(t))/(max(t) - min(t))  # Normalize
    U = transform2U(t, n, p)
    N = SplineBaseFunction(U)
    L = N(t)
    controlpts = curve_control_points(L, points)
    return SplineCurve(N, controlpts)


def function_control_points(x: Iterable[float], y: Iterable[float], p: Optional[int] = None, n: Optional[int] = None):
    ubar = (x-x[0])/(x[-1]-x[0])
    U = transform2U(ubar, n, p)
    N = SplineBaseFunction(U)
    L = N(ubar)
    controlpts = compute_control_points(L, y)
    return controlpts


def compute_control_points(L, y): 
    return la.solve(L @ L.T, L @ y)


from typing import Iterable, Tuple, Union

import numpy as np


class VerifyKnotVector(object):

    minU = 0
    maxU = 1

    @staticmethod
    def isFloatArray1D(U: Tuple[float]) -> None:
        try:
            Ud = np.array(U, dtype="float16")
        except TypeError as e:
            raise TypeError(
                f"Cannot convert U (type={type(U)}) into a numpy float array"
            )
        except ValueError as e:
            raise TypeError(f"All the elements inside U must be floats!")
        if Ud.ndim == 0:
            raise TypeError(
                f"Received U is type {type(U)}, but it's required a Tuple[float]"
            )
        if Ud.ndim != 1:
            raise ValueError(f"U is not a 1D array")

    @staticmethod
    def isOrdenedVector(U: Tuple[float]) -> None:
        n = len(U)
        for i in range(n - 1):
            if U[i] > U[i + 1]:
                raise ValueError("The given U must be ordened")

    @staticmethod
    def Limits(U: Tuple[float]) -> None:
        if VerifyKnotVector.minU is not None:
            VerifyKnotVector.InferiorLimit(U)
        if VerifyKnotVector.maxU is not None:
            VerifyKnotVector.SuperiorLimit(U)

    @staticmethod
    def InferiorLimit(U: Tuple[float]) -> None:
        for u in U:
            if u < VerifyKnotVector.minU:
                raise ValueError(
                    f"All the values in U must be bigger than {VerifyKnotVector.minU}"
                )

    @staticmethod
    def SuperiorLimit(U: Tuple[float]) -> None:
        for u in U:
            if u > VerifyKnotVector.maxU:
                raise ValueError(
                    f"All the values in U must be less than {VerifyKnotVector.maxU}"
                )

    @staticmethod
    def SameQuantityBoundary(U: Tuple[float]) -> None:
        U = np.array(U)
        minU = np.min(U)
        maxU = np.max(U)
        if np.sum(U == minU) != np.sum(U == maxU):
            raise ValueError("U must contain the same quantity of 0 and 1. U = ", U)

    @staticmethod
    def isInteger(val: int):
        if not isinstance(val, int):
            try:
                int(val)
            except Exception as e:
                raise TypeError(f"The value must be an integer. Received {type(val)}")

    @staticmethod
    def isNonNegative(val: int):
        if val < 0:
            raise ValueError(f"The value must be > 0. Received {val}")

    @staticmethod
    def isIntegerNonNegative(value: int):
        VerifyKnotVector.isInteger(value)
        VerifyKnotVector.isNonNegative(value)

    @staticmethod
    def PN(p: int, n: int):
        VerifyKnotVector.isIntegerNonNegative(p)
        VerifyKnotVector.isIntegerNonNegative(n)
        if n <= p:
            raise ValueError("Must n > p. Received n=%d, p=%d" % (n, p))

    @staticmethod
    def all(U: Tuple[float]) -> None:
        VerifyKnotVector.isFloatArray1D(U)
        VerifyKnotVector.isOrdenedVector(U)
        VerifyKnotVector.Limits(U)
        VerifyKnotVector.SameQuantityBoundary(U)


class KnotVector(list):
    def __init__(self, U: Iterable[float]):
        VerifyKnotVector.all(U)
        super().__init__(U)
        self.compute_np()

    @property
    def p(self):
        return self.__p

    @property
    def n(self):
        return self.__n

    @p.setter
    def p(self, value: int):
        VerifyKnotVector.isIntegerNonNegative(value)
        self.__p = int(value)

    @n.setter
    def n(self, value: int):
        VerifyKnotVector.isIntegerNonNegative(value)
        VerifyKnotVector.PN(self.p, value)
        self.__n = int(value)

    def compute_np(self):
        """
        We have that U = [0, ..., 0, ?, ..., ?, 1, ..., 1]
        And that U[p] = 0, but U[p+1] != 0
        The same way, U[n] = 1, but U[n-1] != 0

        Using that, we know that
            len(U) = m + 1 = n + p + 1
        That means that
            m = n + p
        """
        minU = min(self)
        p = 0
        while self[p + 1] == minU:
            p += 1
        self.p = p
        self.n = len(self) - p - 1

    def __find_spot_onevalue(self, u: float) -> int:
        U = np.array(self)
        minU = np.min(self)
        maxU = np.max(self)
        lower = int(np.max(np.where(U == minU)))
        upper = int(np.min(np.where(U == maxU)))
        if u == minU:
            return lower
        if u == maxU:
            return upper
        mid = (lower + upper) // 2
        while True:
            if u < U[mid]:
                upper = mid
            elif U[mid + 1] <= u:
                lower = mid
            else:
                return mid
            mid = (lower + upper) // 2

    def __find_spot_vector(self, u: Iterable[float]) -> np.ndarray:
        """
        find the spots of an vector. It's not very efficient, but ok for the moment
        """
        spots = np.zeros(len(u), dtype="int64")
        for i, ui in enumerate(u):
            spots[i] = self.__find_spot_onevalue(ui)
        return spots

    def spot(self, u: Union[float, Iterable[float]]) -> Union[int, np.ndarray]:
        if isinstance(u, np.ndarray):
            return self.__find_spot_vector(u)
        else:
            return np.array(self.__find_spot_onevalue(u))

    def __eq__(self, __obj: object):
        if not isinstance(__obj, (list, tuple, self.__class__)):
            raise TypeError(
                f"Cannot compare {type(__obj)} with a {self.__class__} instance"
            )
        try:
            __obj = self.__class__(__obj)
        except Exception as e:
            raise ValueError(
                f"No sucess trying to convert {type(__obj)} into {self.__class__}. Cause {str(e)}"
            )
        if self.n != __obj.n:
            return False
        if self.p != __obj.p:
            return False
        for i, v in enumerate(self):
            if v != __obj[i]:
                return False
        return True

    def __ne__(self, __obj: object) -> bool:
        return not self.__eq__(__obj)


class GeneratorKnotVector:
    @staticmethod
    def bezier(p: int) -> KnotVector:
        VerifyKnotVector.isIntegerNonNegative(p)
        return GeneratorKnotVector.uniform(p=p, n=p + 1)

    @staticmethod
    def weight(p: int, ws: Tuple[float]) -> KnotVector:
        VerifyKnotVector.isFloatArray1D(ws)
        VerifyKnotVector.isIntegerNonNegative(p)
        U = np.cumsum(ws)
        U -= U[0]
        U /= U[-1]
        U *= VerifyKnotVector.maxU - VerifyKnotVector.minU
        U += VerifyKnotVector.minU
        U = p * [0] + list(U) + p * [1]
        return KnotVector(U)

    @staticmethod
    def uniform(p: int, n: int) -> KnotVector:
        VerifyKnotVector.PN(p, n)
        ws = np.ones(n - p + 1)
        return GeneratorKnotVector.weight(p=p, ws=ws)

    @staticmethod
    def random(p: int, n: int) -> KnotVector:
        VerifyKnotVector.PN(p, n)
        ws = np.random.random(n - p + 1)
        return GeneratorKnotVector.weight(p, ws)



import numpy as np
from numpy import linalg as la
from typing import Iterable, Optional, Any
from compmec.nurbs import SplineBaseFunction, SplineCurve
from compmec.nurbs.curves import SplineXYFunction

def transform2U(u: Iterable[float], n: float, p: int, algorithm: int = 1):
    if p < 1:
        raise ValueError(f"p must be > 0. Received {p}")
    if n <= p:
        raise ValueError(f"({n} = n must be greater than p = {p}")
    U = np.zeros(n+p+1)
    if algorithm == 1:
        for j in range(1, n-p):
            U[j+p] = np.sum(u[j:j+p])/p
    else:
        ValueError("Algorithm is nod valid!")
    U[n:] = 1
    return U


def curve_spline(p: int, u: Iterable[Iterable[float]], points: Iterable[Iterable[Any]]):
    p = int(p)
    for i, ui in enumerate(u):
        u[i] = np.array(ui)
    for i, pi in enumerate(points):
        points[i] = np.array(pi)
    number_derivatives = len(u)-1
    ndofs = []
    for ui in u:
        ndofs.append(len(ui))
    totalndofs = np.sum(ndofs)
    if totalndofs <=  p:
        raise ValueError(f"The number of informations {totalndofs} very small to interpolate a curve with p = {p}")
    uequaly = np.linspace(0, 1, totalndofs)
    U = transform2U(uequaly, totalndofs, p)
    N = SplineBaseFunction(U)
    functions = [N]
    for i in range(number_derivatives):
        functions.append(functions[-1].derivate())

    Ls = []
    for ui, function in zip(u, functions):
        Ls.append(function(ui))

    dim = points[0].shape[1]
    M = np.zeros((totalndofs, totalndofs))
    B = np.zeros((totalndofs, dim))
    summe = 0
    for i, ndof in enumerate(ndofs):
        M[summe:summe+ndof, :] = Ls[i].T
        B[summe:summe+ndof, :] = points[i][:, :]
        summe += ndof
    controlpoints = np.linalg.solve(M, B)
    return SplineCurve(N, controlpoints)


def function_spline(x: Iterable[float], y: Iterable[float], p: int):
    ubar = (np.array(x)-min(x))/(max(x)-min(x))
    n = len(ubar)
    U = transform2U(ubar, n, p)
    N = SplineBaseFunction(U)
    L = N(ubar)
    X = curve_control_points(L, x)
    Y = curve_control_points(L, y)
    return SplineXYFunction(N, X, Y)


def curve_control_points(L: np.ndarray, y: Iterable[float]):
    return la.solve(L.T, y)


from compmec.nurbs.knotspace import KnotVector
from compmec.nurbs.basefunctions import BaseFunction
from compmec.nurbs.knotspace import KnotVector
from typing import List, Optional, Dict
import numpy as np


def getMultiplicities(U: KnotVector) -> Dict:
    s = {}
    for i, ui in enumerate(U):
        if ui not in s:
            s[ui] = 0
        s[ui] += 1
    return s


def insert_knot_basefunction(F: BaseFunction, knot: float, times: Optional[int] = 1) -> BaseFunction:
    """
    Receives a base function (like N) and returns a new basefuncion such that has the 
    """
    if not isinstance(F, BaseFunction):
        raise TypeError("F must be a BaseFunction instance")
    knotvector = F.U
    spot = knotvector.spot(knot)
    newU = list(knotvector)
    for i in range(times):
        newU.insert(spot+1, knot)
    return F.__class__(newU)


def insert_knot_controlpoints(F: BaseFunction, P: np.ndarray, knot: float, times: Optional[int]=1) -> np.ndarray:
    """
    Receives a curve and returns the controlpoints of the new curve
        * the new base function and the control points
    """
    if not isinstance(F, BaseFunction):
        raise TypeError("F must be a BaseFunction instance")
    U = KnotVector(F.U)
    if not isinstance(U, KnotVector):
        raise TypeError(f"Curve must be a KnotVector instance. Not {type(U)}")
    if times != 1:
        raise ValueError("For the moment, can insert knot only once at the time")
    n, p = U.n, U.p
    k = U.spot(knot)
    a = np.zeros(p)
    for j in range(p):
        i = k - p + 1 + j
        a[j] = (knot - U[i])/(U[i+p] - U[i])
    Q = np.zeros((n+1, P.shape[1]))
    for i in range(k-p+1): # Eq 5.5 at pag 142
        Q[i] = P[i]
    for i in range(k, n):
        Q[i+1] = P[i]
    for i in range(k-p+1, k+1):
        j = i - k + p - 1
        Q[i] = a[j]*P[i] + (1-a[j])*P[i-1]
    return Q

def remove_knot_basefunction(F: BaseFunction, knot: float, times: Optional[int] = 1) -> BaseFunction:
    if not isinstance(F, BaseFunction):
        raise TypeError("F must be a BaseFunction instance")
    U = list(F.U)
    if knot not in U:
        raise ValueError("The knot {knot} is not inside knot vector {U}")
    if times != 1:
        raise ValueError("Can only remove one knot at the time")
    U.remove(knot)
    newF = F.__class__(U)
    return newF

def remove_knot_controlpoints(F: BaseFunction, P: np.ndarray, knot: float, times: Optional[int]=1) -> np.ndarray:
    if not isinstance(F, BaseFunction):
        raise TypeError("F must be a BaseFunction instance")
    U = list(F.U)
    n, p = F.n, F.p
    Pw = np.copy(P)
    r = U.index(knot)
    s = getMultiplicities(U)
    if knot not in s.keys():
        raise ValueError("The given knot is not in the KnotVector")
    t, newU, newP = RemoveCurveKnot(n, p, U, Pw, knot, r, s[knot], times)
    if t == 0:
        newP = newP.tolist()
        newP.pop(r-p+1)
        newP = np.array(newP)
    else:
        raise ValueError("Cannot knot what happens here. Needs testing")
    return np.array(newP)

def Distance4D(P1, P2):
    return np.linalg.norm(np.array(P1)-np.array(P2))

def RemoveCurveKnot(n, p, U, Pw, u, r, s, num):
    """
    Algorith A5.8 from Nurbs Book at page 185
    Remove knot u (index r) num times.
    Input: n, p, U, Pw, u, r, s, num
    Output: t, nw knots & ctrl pts in U & Pw
    s is the multiplicity of of knot u
    """
    Pw = np.array(Pw)
    TOLERANCE = 1e-7
    m = n + p + 1
    ord = p+1
    fout = (2*r-s-p)/2
    last = r-s
    first = r-p
    temp = np.zeros(Pw.shape)
    for t in range(num):  # This loop is Eq. (5.28)
        off = first - 1  # Diff in index between temp and P
        temp[0] = Pw[off]
        temp[last+1-off] = Pw[last+1]
        i = first
        j = last
        ii = 1
        jj = last - off
        remflag = 0
        while j - i > t:  # Compute new control points for one removal step
            alfi = (u-U[i])/(U[i+ord+t]-U[i])
            alfj = (u-U[j-t])/(U[j+ord]-U[j-t])
            temp[ii] = (Pw[i] - (1-alfi)*temp[ii-1])/alfi
            temp[jj] = (Pw[j] - alfj*temp[jj+1])/(1-alfj)
            i += 1
            ii += 1
            j -= 1
            jj -= 1
        if j - i < t:  # Check if knot removable
            if Distance4D(temp[ii-1], temp[jj+1]) <= TOLERANCE:
                remflag = 1
        else:
            alfi = (u-U[i])/(U[i+ord+t]-U[i])
            if Distance4D(Pw[i], alfi*temp[ii+t+1] + (1-alfi)*temp[ii-1]) <= TOLERANCE:
                remflag = 1
        if remflag == 0:  # Cannot remove any more knots
            break
        else:  # Successful removal. Save new cont. pts.
            i = first
            j = last
            while j - i > t:
                Pw[i] = temp[i-off]
                Pw[j] = temp[j-off]
                i += 1
                j -= 1
        first -= 1
        last += 1
    if t == 0:
        return t, U, Pw
    for k in range(r+1, m+1):
        U[k-t] = U[k]  # Shift knots
    j = fout
    i = j  # Pj thru Pi will be overwritten
    for k in range(1, t):
        if k % 2 == 1:  # k modulo 2
            i += 1
        else:
            j -= 1
    for k in range(i+1, n+1):  # Shift
        Pw[j] = Pw[k]
        j += 1
    return t, U, Pw




from compmec.nurbs.basefunctions import BaseFunction, SplineBaseFunction, RationalBaseFunction
import numpy as np
from typing import Iterable, List, Tuple

class BaseCurve(object):
    def __init__(self, F: BaseFunction, P: List[Tuple[float]]):
        self.F = F
        self.P = P

    def __call__(self, u: Iterable[float]) -> np.ndarray:
        L = self.F(u)
        return L.T @ self.P

    def derivate(self):
        dF = self.F.derivate()
        return self.__class__(dF, self.P)

    @property
    def npts(self):
        return self.P.shape[0]

    @property
    def dim(self):
        if len(self.P.shape) == 1:
            return 1
        return self.P.shape[1]

    @property
    def F(self):
        return self.__F
    
    @property
    def P(self):
        return self.__P

    @property
    def U(self):
        return self.F.U

    @F.setter
    def F(self, value: BaseFunction):
        if not isinstance(value, BaseFunction):
            raise TypeError("It must be a BaseFunction instance")
        self.__F = value

    @P.setter
    def P(self, value: List[Tuple[float]]):
        value = np.array(value, dtype="float64")
        if self.F.n != value.shape[0]:
            raise ValueError(f"The number of control points must be the same of degrees of freedom of BaseFunction. F.n = {self.F.n} != {len(value)} = len(P)")
        self.__P = value

    def __eq__(self, __obj: object) -> bool:
        if not isinstance(__obj, self.__class__):
            raise TypeError(f"Cannot compare a {type(__obj)} object with a {self.__class__} object")
        knots = []
        for ui in self.F.U:
            if ui not in knots:
                knots.append(ui)
        utest = []
        for i in range(len(knots)-1):
            utest += list(np.linspace(knots[i], knots[i+1], 1+self.F.p, endpoint=False))
        utest += [knots[-1]]
        utest = np.array(utest)
        Cusel = self(utest)
        Cuobj = __obj(utest)
        return np.all(np.abs(Cusel - Cuobj) < 1e-9)

    def __ne__(self, __obj: object):
        return not self.__eq__(__obj)

    def __neg__(self):
        return self.__class__(self.F, np.copy(-self.P))

    def __add__(self, __obj: object):
        if not isinstance(__obj, self.__class__):
            raise TypeError(f"Cannot sum a {type(__obj)} object with a {self.__class__} object")
        if self.U != __obj.U:
            raise ValueError("The vectors of curves are not the same!")
        if self.P.shape != __obj.P.shape:
            raise ValueError("The shape of control points are not the same!")
        newP = np.copy(self.P) + __obj.P
        return self.__class__(self.F, newP)

    def __sub__(self, __obj: object):
        if not isinstance(__obj, self.__class__):
            raise TypeError(f"Cannot sum a {type(__obj)} object with a {self.__class__} object")
        return self + (-__obj)


class SplineCurve(BaseCurve):
    def __init__(self, F: SplineBaseFunction, controlpoints: np.ndarray):
        super().__init__(F, controlpoints)


class RationalCurve(BaseCurve):
    def __init__(self, f: RationalBaseFunction, controlpoints: np.ndarray):
        super().__init__(f, controlpoints)


class BaseXYFunction(object):
    def __init__(self, f: BaseFunction, xconpoints: Iterable[float], yconpoints: Iterable[float]):
        self.f = f
        self.Y = np.array(yconpoints)
        
        nsample = 129
        self.usample = np.linspace(0, 1, nsample)
        L = self.f(self.usample)
        self.xsample = L.T @ np.array(xconpoints)
        if self.xsample[0] < self.xsample[-1]: # Always increase
            for i in range(nsample-1):
                if not self.xsample[i] < self.xsample[i+1]:
                    print("self.xsample[%d] = " % i, self.xsample[i])
                    print("self.xsample[%d] = " % (i+1), self.xsample[i+1])
                    print("self.usample[%d] = " % i, self.usample[i])
                    print("self.usample[%d] = " % (i+1), self.usample[i+1])
                    raise ValueError("1: The x values must be increasing or decreasing")
        else: # Always decrease
            for i in range(nsample-1):
                if not self.xsample[i] > self.xsample[i+1]:
                    print("self.xsample[%d] = " % i, self.xsample[i])
                    print("self.xsample[%d] = " % (i+1), self.xsample[i+1])
                    print("self.usample[%d] = " % i, self.usample[i])
                    print("self.usample[%d] = " % (i+1), self.usample[i+1])
                    raise ValueError("2: The x values must be increasing or decreasing")

    def find_indexs(self, x: Iterable[float]) -> np.ndarray:
        ind = np.zeros(len(x), dtype="int32")
        maxxsample = np.max(self.xsample)
        for i, xi in enumerate(x):
            value = np.where((xi < self.xsample[1:])*(self.xsample[:-1] <= xi))[0]
            if len(value) != 0:
                ind[i] = value[0]
            elif xi == maxxsample:
                ind[i] = np.where(xi == self.xsample)[0][0]
        return ind

    def inverse(self, x: Iterable[float]) -> np.ndarray:
        u = np.zeros(len(x))
        ind = self.find_indexs(x)
        for i, xi in enumerate(x):
            if ind[i] == len(self.xsample)-1:
                u[i] = self.usample[ind[i]]
                continue
            ua, ub = self.usample[ind[i]], self.usample[ind[i]+1]
            xa, xb = self.xsample[ind[i]], self.xsample[ind[i]+1]
            u[i] += ub*(xi - xa)/(xb - xa)
            u[i] += ua*(xb - xi)/(xb - xa)
        return u

    def __call__(self, x: Iterable[float]) -> np.ndarray:
        u = self.inverse(x)
        L = self.f(u)
        return L.T @ self.Y
    
class SplineXYFunction(BaseXYFunction):
    def __init__(self, f: SplineBaseFunction, xconpoints: Iterable[float], yconpoints: Iterable[float]):
        super().__init__(f, xconpoints, yconpoints)

class RationalXYFunction(BaseXYFunction):
    def __init__(self, f: RationalBaseFunction, xconpoints: Iterable[float], yconpoints: Iterable[float]):
        super().__init__(f, xconpoints, yconpoints)


from compmec.nurbs.basefunctions import BaseFunction
from compmec.nurbs.knotoperations import insert_knot_basefunction, remove_knot_controlpoints
from typing import Optional
import numpy as np


def degree_elevation_basefunction(F: BaseFunction, times: Optional[int]=1):
    if times != 1:
        raise ValueError("Canont do degree elevation many times at once. Call it many times")
    U = list(F.U)
    knotvalues = []
    for ui in U:
        if ui not in knotvalues:
            knotvalues.append(ui)
    for ui in knotvalues:
        index = U.index(ui)
        U.insert(index, ui)
    newF = F.__class__(U)
    return newF

def degree_elevation_controlpoints(F: BaseFunction, P: np.ndarray, times: Optional[int]=1):
    """
    Once we have F (the current base function) and P (the current control points)
    we can get the new base function newF using the function above
    And then we use the least equares method to find Q such best fits the curve 
    """
    if times != 1:
        raise ValueError("Canont do degree elevation many times at once. Call it many times")
    newF = degree_elevation_basefunction(F)
    m = newF.n
    knots = []
    for i, ui in enumerate(newF.U):
        if ui not in knots:
            knots.append(ui)
    utest = []
    ndiv = max(3, int(2*m/(len(knots)-1)))
    for a, b in zip(knots[:-1], knots[1:]):
        utest += list(np.linspace(a, b, ndiv, endpoint=False))
    utest += [1]
    utest = np.array(utest) 
    L = newF(utest)
    M = L @ L.T
    Cu = F(utest).T @ P
    dim = Cu.shape[1]
    Q = np.zeros((m, dim))
    for i in range(dim):
        Q[:, i] = np.linalg.solve(M, L @ Cu[:, i])
    return Q

def degree_reduction_basefunction(F: BaseFunction, times: Optional[int]=1):
    if times != 1:
        raise ValueError("Canont do degree elevation many times at once. Call it many times")
    U = list(F.U)
    knotvalues = []
    for ui in U:
        if ui not in knotvalues:
            knotvalues.append(ui)
    for ui in knotvalues:
        U.remove(ui)
    newF = F.__class__(U)
    return newF

def degree_reduction_controlpoints(F: BaseFunction, P: np.ndarray, times: Optional[int]=1):
    """
    Once we have F (the current base function) and P (the current control points)
    we can get the new base function newF using the function above
    And then we use the least equares method to find Q such best fits the curve 
    """
    if times != 1:
        raise ValueError("Canont do degree elevation many times at once. Call it many times")
    newF = degree_reduction_basefunction(F)
    m = newF.n
    knots = []
    for i, ui in enumerate(newF.U):
        if ui not in knots:
            knots.append(ui)
    utest = []
    ndiv = max(3, int(2*m/(len(knots)-1)))
    for a, b in zip(knots[:-1], knots[1:]):
        utest += list(np.linspace(a, b, ndiv, endpoint=False))
    utest += [1]
    utest = np.array(utest) 
    L = newF(utest)
    M = L @ L.T
    Cu = F(utest).T @ P
    dim = Cu.shape[1]
    Q = np.zeros((m, dim))
    for i in range(dim):
        Q[:, i] = np.linalg.solve(M, L @ Cu[:, i])
    return Q



import numpy as np
from typing import Iterable, Any, Optional, Union, Tuple, Type
from compmec.nurbs.knotspace import KnotVector 

def N(i: int, j: int, k: int, u: float, U: KnotVector) -> float:
    """
    Returns the value of N_{ij}(u) in the interval [u_{k}, u_{k+1}]
    Remember that N_{i, j}(u) = 0   if  ( u not in [U[i], U[i+j+1]] )
    """
    
    n, p = U.n, U.p
    
    if k < i:
        return 0
    if j == 0:
        if i == k:
            return 1
        if i + 1 == n and k == n:
            return 1
        return 0
    if i + j < k:
        return 0

    if i+j > n+p:
        factor1 = 0
    elif U[i] == U[i+j]:
        factor1 = 0
    else:
        factor1 = (u-U[i])/(U[i+j]-U[i])

    if i+j+1 > n+p:
        factor2 = 0
    elif U[i+j+1] == U[i+1]:
        factor2 = 0
    else:
        factor2 = (U[i+j+1]-u)/(U[i+j+1]-U[i+1])

    result = factor1 * N(i, j-1, k, u, U) + factor2 * N(i+1, j-1, k, u, U)
    return result

def R(i: int, j: int, k: int, u: float, U: KnotVector, w: Iterable[float]) -> float:
    """
    Returns the value of R_{ij}(u) in the interval [u_{k}, u_{k+1}]
    """
    Niju = N(i, j, k, u, U)
    if Niju == 0:
        return 0
    n = len(w)
    soma = 0
    for z in range(n):
        soma += w[z] * N(z, j, k, u, U)
    return w[i] * Niju / soma


class EvaluationClass(object):

    def __init__(self, U: Iterable[float], p: int, tup: Union[None, int, slice, Tuple]= None, A: Optional[np.ndarray]=None):
        self.__U = KnotVector(U)
        self._p = p
        self.__initialize_tup(tup)
        self.A = A

    def __initialize_tup(self, tup):
        if tup is None:
            self.i = None
            self.j = self.p
        elif isinstance(tup, tuple):
            if len(tup) > 2:
                raise IndexError("The dimension of N is maximum 2")
            self.i = tup[0]
            self.j = tup[1]
        else:
            self.i = tup
            self.j = self.p

    @property
    def i(self):
        return self._i
    
    @property
    def j(self):
        return self._j

    @property
    def p(self):
        return self._p

    @property
    def n(self):
        return self.U.n

    @property
    def U(self):
        return self.__U

    @property
    def A(self):
        return self._A

    @i.setter
    def i(self, value: Union[None, int, slice]):
        if value is None:
            self._i = range(0, self.n, 1)
            return
        if isinstance(value, slice):
            start = 0 if value.start is None else value.start
            stop = self.U.n if value.stop is None else value.stop
            step = 1 if value.step is None else value.step
            self._i = range(start, stop, step)
            return
        if isinstance(value, range):
            self._i = value
            return
        try:
            value = int(value)
            if value > self.U.n:
                raise ValueError("The frist value must be <= n = %d" % self.U.n)
            self._i = range(value, value+1, 1)
            return
        except Exception as e:
            raise TypeError(f"Type {type(value)} is incorrect to set i")
        
    @j.setter
    def j(self, value: int):
        try:
            value = int(value)
        except Exception as e:
            raise TypeError(f"Type {type(value)} is incorrect to set j")
        if value < 0:
            raise ValueError("The second value (%d) must be >= 0" % value)
        elif self.p < value:
            raise ValueError("The second value (%d) must be <= p = %d" % (value, self.p))
        self._j = value
        return

    @p.setter
    def p(self, value: int):
        if value < 0:
            raise ValueError("p must be >= 0")
        self._p = int(value)

    @A.setter
    def A(self, value: np.ndarray):
        if value is None:
            self._A = np.eye(self.U.n)
        else:
            value = np.array(value)
            if value.ndim != 2 or value.shape[0] != self.U.n:
                raise ValueError("The given numpy array must be a square matrix")
            self._A = np.array(value)


    def __validate_evaluation__U(self, u: np.ndarray):
        U = self.__U
        minU = np.min(U)
        maxU = np.max(U)
        if not np.all((minU <= u) * (u <= maxU)):
            raise Exception(f"All values of u must be inside the interval [{minU}, {maxU}]")
        if np.array(u).ndim > 1:
            raise ValueError("For the moment we can only evaluate scalars or 1D array")
        

    def __treat_input(self, u: Union[float, np.ndarray]) -> np.ndarray:
        try:
            len(u)
        except Exception as e:
            u = [u]
        return np.array(u)

    def compute_matrix(self, u: Union[float, np.ndarray]) -> np.ndarray:
        """
        In this function, u is a vector, while i is a range
        """
        u = self.__treat_input(u)
        r = np.zeros((self.n, len(u)))
        for w, uw in enumerate(u):
            k = self.__U.spot(uw)
            for ind in self.i:
                r[ind, w] = self.f(ind, self.j, k, uw, self.__U)
        return r
    
    def compute_all(self, u: np.ndarray) -> np.ndarray:
        return self.A @ self.compute_matrix(u)

    def __call__(self, u: Union[float, np.ndarray]) -> np.ndarray:
        self.__validate_evaluation__U(u)
        u = self.__treat_input(u)
        M = self.compute_matrix(u)
        try:
            float(u)
            M = M.reshape(len(M))
        except Exception as e:
            pass
        result = self.A @ M
        return result[self.i]


class BaseFunction(object):

    def __init__(self, U: KnotVector):
        """
        We have that the object can be called like
        N = SplineBaseFunction(U)

        And so, to get the value of N_{i, j}(u) can be express like
        N[i, j](u)
        """
        self.U = KnotVector(U)
        self.p = self.__U.p
        self.A = np.eye(self.n)


    @property
    def p(self) -> int:
        return self.__p

    @property
    def n(self) -> int:
        return self.__U.n

    @property
    def U(self) -> KnotVector:
        return KnotVector(self.__U)

    @property
    def A(self) -> np.ndarray:
        return self.__A

    @p.setter
    def p(self, value: int) -> None:
        if value < 0:
            raise ValueError("Cannot set p = ", value)
        self.__p = int(value)

    @U.setter
    def U(self, value: KnotVector) -> None:
        self.__U = KnotVector(value)

    @A.setter
    def A(self, value: np.ndarray) -> None:
        self.__A = value

    

    @property
    def evaluationClass(self) -> type[EvaluationClass]:
        """
        This function must be overwritten.
        * If it's spline, evalfunction = N
        * If it's rational, evalfunction = R
        """
        raise NotImplementedError("This function must be overwritten")


    def createEvaluationInstance(self, tup: Tuple[slice, int]) -> EvaluationClass:
        return self.evaluationClass(self.U, self.p, tup, self.A)

    def __getitem__(self, tup: slice) -> EvaluationClass:
        return self.createEvaluationInstance(tup)

    def __call__(self, u: np.ndarray) -> np.ndarray:
        return self[:, self.p] (u)

    def derivate(self):
        F = self
        U = list(F.U)
        n, p = F.n, F.p
        A = F.A
        newinstance = F.__class__(U)
        newinstance.p = p - 1
        avals = np.zeros(n)
        for i in range(n):
            diff = U[i+p] - U[i]
            if diff != 0:
                avals[i] = p/diff
        newA = np.diag(avals)
        for i in range(n-1):
            newA[i, i+1] = -avals[i+1]
        newinstance.A = A @ newA
        return newinstance

    def __eq__(self, value):
        if not isinstance(value, self.__class__):
            raise TypeError(f"Cannot compare a {self.__class__} instance with a {type(value)}")
        if self.U != value.U:
            return False
        if np.any(self.A != value.A):
            return False
        return True

class SplineBaseFunction(BaseFunction):

    def __doc__(self):
        """
        This function is recursively determined like

        N_{i, 0}(u) = { 1   if  U[i] <= u < U[i+1]
                      { 0   else

                          u - U[i]
        N_{i, j}(u) = --------------- * N_{i, j-1}(u)
                       U[i+j] - U[i]
                            U[i+j+1] - u
                      + ------------------- * N_{i+1, j-1}(u)
                         U[i+j+1] - U[i+1]

        As consequence, we have that

        N_{i, j}(u) = 0   if  ( u not in [U[i], U[i+j+1]] )

        """
    
    def __init__(self, U: Iterable[float]):
        super().__init__(U)

    @property
    def evaluationClass(self) -> type[EvaluationClass]:
        return SplineEvaluationClass

class RationalBaseFunction(BaseFunction):
    def __init__(self, U: Iterable[float]):
        super().__init__(U)
        self.w = np.ones(self.n)

    @property
    def evaluationClass(self) -> type[EvaluationClass]:
        return RationalEvaluationClass

    @property
    def w(self):
        return self.__w

    @w.setter
    def w(self, value: Iterable[float]):
        for v in value:
            v = float(v)
            if v < 0:
                raise ValueError("The weights must be positive")
        self.__w = value
    
    def __eq__(self, value):
        if not super().__eq__(value):
            return False
        if self.w != value.w:
            return False
        return True


class SplineEvaluationClass(EvaluationClass):

    def __init__(self, U: Iterable[float], p: int, tup: Any, A: Optional[np.ndarray]=None):
        super().__init__(U, p, tup, A)
        
    @property
    def f(self):
        return N


class RationalEvaluationClass(EvaluationClass):
    def __init__(self, U: Iterable[float], w: Iterable[float], p: int, tup: Any, A: Optional[np.ndarray]=None):
        super().__init__(U, p, tup, A)
        self.w = w
        
    @property
    def f(self):
        raise NotImplementedError("TO DO")
        return R

    @property
    def w(self):
        return self._w

    @w.setter
    def w(self, value: Iterable[float]):
        value = np.array(value)
        if value.ndim != 1:
            raise ValueError("to set the weights, it must be a vector")
        if len(value) != self.U.n:
            raise ValueError("The size of weights must be the same as number of functions")
        self._w = value

    


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nurbs']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'compmec-nurbs',
    'version': '0.2.3',
    'description': '',
    'long_description': "[![PyPI Version][pypi-image]][pypi-url]\n[![Build Status][build-image]][build-url]\n[![Code Coverage][coverage-image]][coverage-url]\n[![][versions-image]][versions-url]\n\n# Nurbs\n\nThis repository contains code for inteporlate functions using B-Splines and Nurbs.\n\n\n#### Features\n\n* Basic Functions\n    * Spline ```N```\n    * Rational ```R```\n    * Derivative\n* Curves\n    * Spline\n    * Rational\n* Knot operations\n    * Insertion\n    * Removal\n* Degree operations\n    * Degree elevation\n    * Degree reduction\n\n## Install\n\nThis library is available in [PyPI][pypilink]. To install it\n\n```\npip install compmec-nurbs\n```\n\n## Documentation\n\nIn progress\n\n\n# FAQ\n\n#### Must I learn the theory to use it?\n\nNo! Just see the examples and it will be fine\n\n#### Can I understand the code here?\n\nYes! The easier way is to look up the **python notebook** which contains the theory along examples\n\n#### Is this code efficient?\n\nNo. It's written in python and the functions were made for easy understanding, not for performance.\nThat means: It's not very fast, but it works fine.\n\n\n## Contribute\n\nPlease use the [Issues][issueslink] or refer to the email ```compmecgit@gmail.com```\n\n<!-- Badges: -->\n\n[pypi-image]: https://img.shields.io/pypi/v/compmec-nurbs\n[pypi-url]: https://pypi.org/project/compmec-nurbs/\n[build-image]: https://github.com/compmec/nurbs/actions/workflows/build.yaml/badge.svg\n[build-url]: https://github.com/compmec/nurbs/actions/workflows/build.yaml\n[coverage-image]: https://codecov.io/gh/compmec/nurbs/branch/main/graph/badge.svg\n[coverage-url]: https://codecov.io/gh/compmec/nurbs/\n[versions-image]: https://img.shields.io/pypi/pyversions/compmec-nurbs.svg?style=flat-square\n[versions-url]: https://pypi.org/project/compmec-nurbs/\n[pypilink]: https://pypi.org/project/compmec-nurbs/\n[issueslink]: https://github.com/compmec/nurbs/issues\n",
    'author': 'Carlos Adir',
    'author_email': 'carlos.adir@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
